/**
 * Retail Data Integration Services
 * (C) Copyright IBM Corp. 2003
 */

package examples.soap;

import java.net.URL;

import javax.xml.soap.*;

import org.apache.commons.logging.*;

import com.ibm.retail.di.naming.Context;
import com.ibm.retail.di.net.protocol.MessageHeaders;
import com.ibm.retail.di.xml.messaging.service.AbstractActor;
import com.ibm.retail.di.xml.messaging.soap.SoapUtil;


/**
 * This Actor establishes a connection to a remote SOAP service and sends
 * the inbound SOAP message request, then passes back the response. The target
 * service is determined by the message profile Id of the inbound request. The
 * Actor retrieves the target URL from the Context; it must have the form
 * <code><i>ActorName</i>.<i>Message-Profile-Id</code>.url=<i>url</i>. If the
 * key is not defined, an exception is thrown and a fault is generated for
 * the request. 
 * <p>
 * Note that this Actor expects HTTP or SOEP URLs only. Any other protocol
 * will require a custom <code>SOAPConnectionFactory</code> to handle the 
 * connection. This can be accomplished in <code>createSoapConnection</code>.
 * </p>
 * 
 * @version 1.0
 * @author C. Capps
 */
public class RemoteSoapConnectionActor extends AbstractActor {

    // IBM copyright
    static String COPYRIGHT = com.ibm.retail.di.Copyright.IBM_COPYRIGHT_SHORT;

    
    /** creates a Log4J logger for use by this class */    
    static Log logger = LogFactory.getLog(RemoteSoapConnectionActor.class);
    
    
    /**
     * Determines the remote SOAP URL to call for the inbound request, 
     * establishes the connection, sends the request, and returns the
     * response. If the target URL is not defined or cannot be reached,
     * a fault is generated.
     * <p>
     * This method calls <code>getUrlEndpoint</code> to determine the
     * target SOAP service and uses <code>createSoapConnection</code>
     * to build the connection object.
     * </p>
     * 
     * @param aMessage the inbound SOAP request
     * @return the result of calling the remote SOAP service
     */
    public SOAPMessage onMessage(SOAPMessage aMessage) {
        SOAPMessage response = null;
        
        // what is the profile id of this message?
        String mpId = SoapUtil.getMessageProfileId(aMessage);
        
        // what is the endpoint for this message?
        URL urlEndpoint = null;
        try {
            urlEndpoint = getUrlEndpoint(mpId);
            SOAPConnection soapConn = createSoapConnection(urlEndpoint);
            
            if (logger.isDebugEnabled()) {
                logger.debug("remote: calling soap connection endpoint [" + urlEndpoint + "]");
            }
            
            response = soapConn.call(aMessage, urlEndpoint);
            
            if (logger.isDebugEnabled()) {
                logger.debug("remote: call complete [" + urlEndpoint + "]");
            }            
            
            //response.writeTo(System.out);
            
        } catch (Exception e) {
            
            logger.warn(e);
            response = createSoapFaultMessage(e);            
        }
        
        return response;
    }
    
    
    /**
     * Determines the URL with which the message profile Id argument is
     * associated. Messages having this profile will be sent to the URL
     * returned by this method. By default, the associated URL is
     * retrieved from the Context; see above for how the key should
     * be formed.
     * 
     * @param aProfileId a message profile Id
     * @return a URL instance associated with this message profile Id
     * @throws Exception If the endpoint cannot be determined, such as
     *          when the required Context key is not defined
     */
    protected URL getUrlEndpoint(String aProfileId) throws Exception {
        String endpointUrlStr = getContextString(getEndpointContextProperty(aProfileId), Context.REQUIRED);
        URL endpointUrl = new URL(endpointUrlStr);
        return endpointUrl;
    }

        
    /**
     * Returns the name of the Context property used to retrieve the
     * URL String associated with the message profile Id argument.
     * 
     * @param aProfileId a message profile Id
     * @return the name of the Context property used to retrieve the
     *          associated URL String
     */
    protected String getEndpointContextProperty(String aProfileId) {
        return aProfileId + ".url";
    }
    
    
    /**
     * Creates and returns a SOAP connection to the given URL endpoint. By
     * default, this method only handles HTTP or SOEP protocols.
     * 
     * @param aUrlEndpoint the URL describing the remote SOAP service
     * @return a SOAP connection to the given URL endpoint
     * @throws Exception If the connection cannot be established or the
     *          protocol is not supported
     */
    protected SOAPConnection createSoapConnection(URL aUrlEndpoint) throws Exception {
        SOAPConnectionFactory factory = null;
        
        if (aUrlEndpoint.getProtocol().equalsIgnoreCase("http")) {
            factory = SoapUtil.getHttpSoapConnectionFactory();
        } else {
            factory = SoapUtil.getDefaultSoapConnectionFactory();
        }
        
        return factory.createConnection();            
    }
}