/**
 * Retail Data Integration Services
 * (C) Copyright IBM Corp. 2003
 */

package examples.soap;

import java.io.InputStream;

import javax.xml.soap.SOAPMessage;

import org.apache.commons.logging.*;

import com.ibm.retail.RetailException;
import com.ibm.retail.di.net.protocol.MessageHeaders;
import com.ibm.retail.di.util.ByteTools;
import com.ibm.retail.di.xml.messaging.service.AbstractActor;
import com.ibm.retail.di.xml.messaging.soap.SoapUtil;


/**
 * This actor ensures that its output is a simple byte array rather
 * than a SOAP message. If the input is a byte array, it is returned
 * unchanged; if the input is a SOAP message, it is converted to a byte
 * array and then returned. The conversion should be defined by 
 * subclasses in <code>getRequestAsStream</code>. By default, the SOAP
 * message is converted according to the base class behavior for that
 * method; see below.
 * <p>
 * <i>Common Use</i><br>
 * This class is useful as the last Actor in a series whenever
 * the messaging style is request/response and the source is a
 * non-Java or non-SOAP application (such as a C or CBASIC application).
 * The actor ensures that the response will be converted to bytes
 * before being handed back to the calling application. This is only
 * necessary if the response might be a SOAP message.
 * </p>
 * 
 * @author C. Capps
 */
public class SoapStreamConversionActor extends AbstractActor {


    // IBM copyright
    static String COPYRIGHT = com.ibm.retail.di.Copyright.IBM_COPYRIGHT_SHORT;

    
    /** creates a Log4J logger for use by this class */    
    static Log logger = LogFactory.getLog(SoapStreamConversionActor.class);

	/**
	 * Generates a response for the request argument. The request is accompanied by a
	 * collection of request headers. The request itself is an <code>Object</code>.
	 * <P>
	 * In this implementation, if the request is deemed a SOAP message, the message
	 * is converted to a byte array, which is returned. If the request is non-SOAP,
	 * it is converted to a SOAPMessage, which is returned. The conversion process
	 * is handled by subclasses.
	 * </p>
	 * 
	 * @param theRequestHeaders
	 *                   the headers in the inbound request
	 * 
	 * @param theRequest the data of the request message, which must be
	 *                   a byte array or input stream
	 * 
	 * @param theResponseHeaders
	 *                   an empty <CODE>MessageHeaders</CODE> instance whose
	 *                   contents will be populated with response headers
	 * 
	 * @return a byte array containing the content of the response
	 */
	public Object onMessage(MessageHeaders theRequestHeaders,
							Object theRequest,
							MessageHeaders theResponseHeaders) {

		Object response = null;
		boolean isSoapRequest = 
			isSoapMessage(theRequestHeaders, theRequest);

		try {
			
			// If SOAP is input, we output bytes
			if (isSoapRequest) {
				
				if (logger.isDebugEnabled()) {
					logger.debug("inbound message is soap: converting to bytes");
				}
				SOAPMessage soapMessage = getSoapMessage(theRequestHeaders, theRequest);
				byte[] b = convertSoapMessageToBytes(theRequestHeaders, soapMessage, theResponseHeaders);
				
				response = b;
				
			} else {

				if (logger.isDebugEnabled()) {
					logger.debug("inbound message is bytes: converting to soap");
				}				
				// if input is byte[], output SOAP
				byte[] b = ByteTools.getObjectAsBytes(theRequest);
				SOAPMessage soapMessage = convertBytesToSoapMessage(theRequestHeaders, b, theResponseHeaders);
				
				response = soapMessage;
			}
		}
		catch (Exception e) {
			response = createSoapFaultMessage(e, theResponseHeaders);
		}
		
		return response;
	}


	// users provide their own impl
    protected byte[] convertSoapMessageToBytes(MessageHeaders theRequestHeaders, SOAPMessage soapMessage, MessageHeaders theResponseHeaders) throws Exception
    {
    	// get all the attachments of the inbound SOAP message, converted to bytes and concatenated
        byte[] b = SoapUtil.getAttachmentContentAsBytes(soapMessage);
        return b;
    }


	//	users provide their own impl
    protected SOAPMessage convertBytesToSoapMessage(MessageHeaders theRequestHeaders, byte[] b, MessageHeaders theResponseHeaders) throws Exception
    {
    	// create a SOAP message with a single binary attachment
        SOAPMessage request = SoapUtil.createSoapMessage(getMessageFactory(), null, b);
        return request;
    }


    public SOAPMessage onMessage(SOAPMessage aMessage)
    {
		// never should be called by this Actor, since SOAP processing always results in a byte array response
        SOAPMessage soapMessage = createSoapFaultMessage(
        	new RetailException("Bug! onMessage() called in convertSoapToBytes"),
        	null);
        	
        return soapMessage;
    }
}
