package examples.syncpointed;

/**
 * IBM Confidential                                                  
 * OCO Source Materials                                              
 * Satin Data Integration Toolkit
 * Prog License number                                               
 * (C) Copyright IBM Corp. 2002                                      
 * The source code for this program is not published or              
 * otherwise divested of its trade secrets, irrespective of          
 * what has been deposited with the US Copyright Office.             
 */

import java.io.*;

import javax.xml.soap.*;

import com.ibm.mqe.*;
import com.ibm.mqe.mqemqmessage.*;

import com.ibm.retail.di.net.protocol.*;
import com.ibm.retail.di.net.service.interactive.*;
import com.ibm.retail.di.service.util.*;
import com.ibm.retail.di.service.mom.mqe.*;
import com.ibm.retail.di.xml.messaging.soap.*;

import org.apache.commons.logging.*;

/**
 * Syncpointed service that distributes messages to actor(s) and then confirms or undoes them.
 * <p>
 * This listens to a MQe queue and sends each message to its Director, which sends the message to
 * its Actors.  If a fault occurs, the message is saved to a faults directory; otherwise the message
 * is confirmed, which removes it from the queue.
 * @author Dan Streetman
 */
public class MqeSyncpointedService extends AbstractInteractiveService implements MQeMessageListenerInterface
{
	/** Initialize - register with MqeInterface and get MqeQueueManager and queue name */
	protected void doInit() throws Exception
	{
                mI = MqeInterfaceFactory.createMqeInterface(getContext());
		mI.registerUser(this);

		super.doInit();
	}

	/** Start - add this as a message listener */
	protected void doStart() throws Exception
	{
		faultDirectory = NamingTools.lookupWithHeaders(getContext(), null, getName(), FAULT_DIRECTORY_KEY, true);
		profileId = NamingTools.lookupWithHeaders(getContext(), null, getName(), PROFILE_CONTEXT_KEY, true);
		queueName = NamingTools.lookupWithHeaders(getContext(), null, getName(), QUEUE_CONTEXT_KEY, true);
		queueManager = mI.getQueueManager();
		queueManager.addMessageListener(this, queueName, null);

		/* process what's already in the queue in a separate Thread */
		Runnable r = new Runnable() { public void run() { processQueue(); } };
		Thread t = new Thread(r);
		t.setDaemon(true);
		t.start();
	}

	/** Stop - remove this as a message listener */
	public void doStop() throws Exception
	{
		synchronized (mI) {
			if (!mI.isClosed())
				mI.getQueueManager().removeMessageListener(this, queueName, null);
		}
	}

	/** Close - close the MqeInterface */
	public void close()
	{
		try { mI.close(this); }
		catch ( Exception e ) { logger.warn("Exception during close : " + e.toString()); }

		super.close();
	}

	/**
	 * Process messages in the incoming queue until empty.
	 */
	public void processQueue()
	{
		boolean queueEmpty = false;

		do {
			try {
				queueEmpty = !processNextMessage();
			} catch ( Exception e ) {
				queueEmpty = false;
			}
		} while (!queueEmpty);
	}

	/**
	 * Process the next message (if any) in the queue.
	 * <p>
	 * This gets the next message using a unique confirmId, sends it
	 * to the listener's onMessage.  Any fault or exception will cause
	 * the message to be placed in a fault-message directory.
	 * @return True if a message was processed, false if the queue was empty.
	 * @exception Exception If an error occurs before getMessage.
	 */
	public boolean processNextMessage() throws Exception
	{
		long confirmId = MQe.uniqueValue();
		MQeMsgObject msg = null;

		try {
			/* Note that if the JVM crashes after getting the message off the queue,
			 * that message will be "stuck" on the queue indefinately.  A truly assured
			 * service must write the confirmId to a persistent location (such as a file)
			 * before getting the message using that confirmId.  Then, if the JVM crashes
			 * this program can read the persistent data on startup, and handle any
			 * unconfirmed messages on the queue.
			 */
			logger.debug("Getting message from queue " + queueName + " using confirmId " + confirmId);
			msg = queueManager.getMessage(null, queueName, null, null, confirmId);
			logger.debug("Got message from queue " + queueName + " using confirmId " + confirmId);
		} catch ( MQeException mE ) {
			if (MQe.Except_Q_NoMatchingMsg == mE.code()) {
				logger.debug("No more messages in queue.");
				return false;
			} else {
				logger.error("MQeException code " + mE.code() + " while getting message : " + mE.toString());
				logger.trace(throwableToString(mE));
				throw mE;
			}
		} catch ( Exception e ) {
			logger.error("Exception while getting message : " + e.toString());
			logger.trace(throwableToString(e));
			throw e;
		}

		try {
			MQeMQMsgObject mqmsg = (MQeMQMsgObject)msg;
			byte[] data = mqmsg.getData();
			Object payload = null;

			if (SoapUtil.isSoapStream(data)) {
				logger.debug("Message is SoapStream");
				payload = SoapUtil.restoreSoapMessage(MessageFactory.newInstance(), data);
			} else {
				logger.debug("Message is not SoapStream");
				payload = data;
			}

			MessageHeaders requestHeaders = new MessageHeaders();
			MessageHeaders responseHeaders = new MessageHeaders();
			requestHeaders.addHeader(MessageHeaders.getMessageProfileIdFieldName(), profileId);

			logger.debug("Sending message with confirmId " + confirmId + " and profileId " + profileId + " to listener.");
			doInteraction(requestHeaders, payload, responseHeaders);
			logger.debug("Finished listener.onMessage for message with confirmId " + confirmId + " and profileId " + profileId);

			int faultCount = Integer.decode(responseHeaders.getHeaderField(MessageHeaders.getFaultCountFieldName())).intValue();

			if (0 < faultCount) {
				logger.debug(faultCount + " faults occured for confirmId " + confirmId);
				writeFaultMessage(confirmId, msg);
			} else {
				logger.debug("No faults for confirmId " + confirmId);
			}
		} catch ( SOAPException sE ) {
			logger.error("SOAPException while processing message : " + sE.toString());
			logger.trace(throwableToString(sE));
			writeFaultMessage(confirmId, msg);
		} catch ( ClassCastException ccE ) {
			logger.error("MQeMsgObject is not MQeMQMsgObject : " + ccE.toString());
			writeFaultMessage(confirmId, msg);
		} catch ( Exception e ) {
			logger.error("Exception processing message : " + e.toString());
			logger.trace(throwableToString(e));
			writeFaultMessage(confirmId, msg);
		}

		logger.debug("Removing message from queue " + queueName);
		try {
			queueManager.confirmGetMessage(null, queueName, msg.getMsgUIDFields());
		} catch ( Exception e ) {
			logger.error("Exception while confirming message on queue " + queueName + " : " + e.toString());
			logger.trace(throwableToString(e));
			logger.error("Leaving unconfirmed messsage on queue!");
		}

		return true;
	}

	/**
	 * Write the message to a file.
	 * @param confirmId The confirmId of the message.
	 * @param message The message.
	 */
	public void writeFaultMessage(long confirmId, MQeMsgObject message)
	{
		String fileName = faultDirectory + File.separator + String.valueOf(confirmId);
		try {
			FileOutputStream foS = new FileOutputStream(fileName);

			foS.write(message.dump());

			foS.close();
		} catch ( Exception e ) {
			logger.error("Exception while writing fault message to file " + fileName + " : " + e.toString());
			logger.trace(throwableToString(e));
			logger.error("Dropping fault message!");
		}
	}

	/** Process a new message */
	public void messageArrived(MQeMessageEvent mmE)
	{
		logger.debug("MQeMessageEvent arrived for queue " + mmE.getQueueName());
		
		while (true) {
			try {
				processNextMessage();
				break;
			} catch ( Exception e ) { }
		}
	}

	/**
	 * Create a String containing a Throwable's stack trace.
	 * @param throwable The Throwable.
	 * @return A String of the Throwable's stack trace.
	 */
	protected String throwableToString(Throwable throwable)
	{
		StringWriter stringWriter = new StringWriter();
		throwable.printStackTrace(new PrintWriter(stringWriter));
		return stringWriter.toString();
	}

	protected String queueName = null;
	protected String faultDirectory = null;
	protected String profileId = null;
	protected MQeQueueManager queueManager = null;
        protected MqeInterface mI = null;

	static Log logger = LogFactory.getLog(MqeSyncpointedService.class);

	public static final String QUEUE_CONTEXT_KEY = "queue";
	public static final String PROFILE_CONTEXT_KEY = "message-profile-id";
	public static final String FAULT_DIRECTORY_KEY = "fault-directory";

}
